import json
import logging
import warnings
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    TypedDict,
    Union,
    cast,
    runtime_checkable,
)

from freeplay.errors import (
    FreeplayClientError,
    FreeplayConfigurationError,
    log_freeplay_client_warning,
)
from freeplay.llm_parameters import LLMParameters
from freeplay.model import (
    InputVariables,
    MediaInputMap,
    MediaInputUrl,
    NormalizedOutputSchema,
)
from freeplay.resources.adapters import (
    MediaContentBase64,
    MediaContentUrl,
    MissingFlavorError,
    TextContent,
    adaptor_for_flavor,
)
from freeplay.support import (
    CallSupport,
    HistoryTemplateMessage,
    MediaSlot,
    MediaType,
    PromptTemplate,
    PromptTemplateMetadata,
    PromptTemplates,
    Role,
    TemplateChatMessage,
    TemplateMessage,
    ToolSchema,
    TemplateVersionResponse,
)
from freeplay.utils import bind_template_variables, convert_provider_message_to_dict

logger = logging.getLogger(__name__)


class UnsupportedToolSchemaError(FreeplayConfigurationError):
    def __init__(self) -> None:
        super().__init__("Tool schema not supported for this model and provider.")


class UnsupportedOutputSchema(FreeplayConfigurationError):
    def __init__(self) -> None:
        super().__init__(
            "Structured outputs are not supported for this model and provider."
        )


class VertexAIToolSchemaError(FreeplayConfigurationError):
    def __init__(self) -> None:
        super().__init__(
            "Vertex AI SDK not found. Install google-cloud-aiplatform to get proper Tool objects."
        )


# Models ==


# A content block compatible with stainless generated SDKs (such as Anthropic
# and OpenAI). This lets us generate a dictionary from the stainless classes
# correctly. Intentionally over-permissive to allow schema evolution by the
# providers.
@runtime_checkable
class ProviderMessageProtocol(Protocol):
    def model_dump(self) -> Dict[str, Any]: ...


class MessageDict(TypedDict):
    role: str
    content: Any


# This type represents a struct or dict containing a role and content. The role
#  should be one of user, assistant or system. This type should be compatible
#  with OpenAI and Anthropic's message format, as well as most other SDKs. If
#  not using a common provider, use {'content': str, 'role': str} to record. If
#  using a common provider, this is usually the `.content` field.
ProviderMessage = Union[MessageDict, Dict[str, Any], ProviderMessageProtocol]

# DEPRECATED: Use ProviderMessage instead
GenericProviderMessage = ProviderMessage


# SDK-Exposed Classes


@dataclass
class PromptVersionInfo:
    prompt_template_version_id: str
    environment: Optional[str]


@dataclass
class PromptInfo(PromptVersionInfo):
    prompt_template_id: str
    prompt_template_version_id: str
    template_name: str
    model_parameters: LLMParameters
    provider_info: Optional[Dict[str, Any]]
    provider: str
    model: str
    flavor_name: str


class FormattedPrompt:
    def __init__(
        self,
        prompt_info: PromptInfo,
        messages: List[Dict[str, str]],
        formatted_prompt: Optional[List[Dict[str, str]]] = None,
        formatted_prompt_text: Optional[str] = None,
        tool_schema: Optional[List[Dict[str, Any]]] = None,
        formatted_output_schema: Optional[Dict[str, Any]] = None,
    ):
        # These two definitions allow us to operate on typed fields until we expose them as Any for client use.
        self._llm_prompt = formatted_prompt
        self._tool_schema = tool_schema
        self._formatted_output_schema = formatted_output_schema

        self.prompt_info = prompt_info
        if formatted_prompt_text:
            self.llm_prompt_text = formatted_prompt_text

        maybe_system_content = next(
            (message["content"] for message in messages if message["role"] == "system"),
            None,
        )
        self.system_content = maybe_system_content

        self._messages = messages

    @property
    def messages(self) -> List[Dict[str, str]]:
        warnings.warn(
            "The 'messages' attribute is deprecated and will be removed in a future version. It is not formatted for the provider. Use 'llm_prompt' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._messages

    @property
    def llm_prompt(self) -> Any:
        return self._llm_prompt

    @property
    def tool_schema(self) -> Any:
        return self._tool_schema

    @property
    def formatted_output_schema(self) -> Any:
        return self._formatted_output_schema

    def all_messages(self, new_message: ProviderMessage) -> List[Dict[str, Any]]:
        converted_message = convert_provider_message_to_dict(new_message)
        return self._messages + [converted_message]


class BoundPrompt:
    def __init__(
        self,
        prompt_info: PromptInfo,
        messages: List[Dict[str, Any]],
        tool_schema: Optional[List[ToolSchema]] = None,
        output_schema: Optional[NormalizedOutputSchema] = None,
    ):
        self.prompt_info = prompt_info
        self.messages = messages
        self.tool_schema = tool_schema
        self.output_schema = output_schema

    @staticmethod
    def __format_tool_schema(flavor_name: str, tool_schema: List[ToolSchema]) -> Any:
        if flavor_name == "anthropic_chat":
            return [
                {
                    "name": tool_schema.name,
                    "description": tool_schema.description,
                    "input_schema": tool_schema.parameters,
                }
                for tool_schema in tool_schema
            ]
        elif flavor_name in ["openai_chat", "azure_openai_chat"]:
            return [
                {"function": asdict(tool_schema), "type": "function"}
                for tool_schema in tool_schema
            ]
        elif flavor_name == "amazon_bedrock_converse":
            return {
                "tools": [
                    {
                        "toolSpec": {
                            "name": tool_schema.name,
                            "description": tool_schema.description,
                            "inputSchema": {"json": tool_schema.parameters},
                        }
                    }
                    for tool_schema in tool_schema
                ]
            }
        elif flavor_name == "gemini_chat":
            try:
                from vertexai.generative_models import Tool, FunctionDeclaration  # type: ignore[import-untyped]

                function_declarations = [
                    FunctionDeclaration(
                        name=tool_schema.name,
                        description=tool_schema.description,
                        parameters=tool_schema.parameters,
                    )
                    for tool_schema in tool_schema
                ]
                return [Tool(function_declarations=function_declarations)]
            except ImportError:
                raise VertexAIToolSchemaError()

        raise UnsupportedToolSchemaError()

    @staticmethod
    def __format_output_schema(
        flavor_name: str, output_schema: NormalizedOutputSchema
    ) -> Dict[str, Any]:
        # For OpenAI and Azure OpenAI, the normalized format is compatible with the API format
        if flavor_name in ["openai_chat", "azure_openai_chat"]:
            return output_schema
        # Add other flavors as necessary - currently only OpenAI-compatible models support output schema
        raise UnsupportedOutputSchema()

    def format(self, flavor_name: Optional[str] = None) -> FormattedPrompt:
        final_flavor = flavor_name or self.prompt_info.flavor_name
        adapter = adaptor_for_flavor(final_flavor)
        formatted_prompt = adapter.to_llm_syntax(self.messages)
        formatted_tool_schema = (
            BoundPrompt.__format_tool_schema(final_flavor, self.tool_schema)
            if self.tool_schema
            else None
        )
        formatted_output_schema = (
            BoundPrompt.__format_output_schema(final_flavor, self.output_schema)
            if self.output_schema
            else None
        )

        if isinstance(formatted_prompt, str):
            return FormattedPrompt(
                prompt_info=self.prompt_info,
                messages=self.messages,
                formatted_prompt_text=formatted_prompt,
                tool_schema=formatted_tool_schema,
                formatted_output_schema=formatted_output_schema,
            )
        else:
            return FormattedPrompt(
                prompt_info=self.prompt_info,
                messages=self.messages,
                formatted_prompt=formatted_prompt,
                tool_schema=formatted_tool_schema,
                formatted_output_schema=formatted_output_schema,
            )


def extract_media_content(
    media_inputs: MediaInputMap, media_slots: List[MediaSlot]
) -> List[Union[MediaContentBase64, MediaContentUrl]]:
    media_content: List[Union[MediaContentBase64, MediaContentUrl]] = []
    for slot in media_slots:
        file = media_inputs.get(slot.placeholder_name, None)
        if file is None:
            continue
        if isinstance(file, MediaInputUrl):
            media_content.append(
                MediaContentUrl(
                    type=slot.type, url=file.url, slot_name=slot.placeholder_name
                )
            )
        else:
            media_content.append(
                MediaContentBase64(
                    type=slot.type,
                    content_type=file.content_type,
                    data=file.data,
                    slot_name=slot.placeholder_name,
                )
            )

    return media_content


class TemplatePrompt:
    def __init__(
        self,
        prompt_info: PromptInfo,
        messages: List[TemplateMessage],
        tool_schema: Optional[List[ToolSchema]] = None,
        output_schema: Optional[NormalizedOutputSchema] = None,
    ):
        self.prompt_info = prompt_info
        self.tool_schema = tool_schema
        self.output_schema = output_schema
        self.messages = messages

    def bind(
        self,
        variables: InputVariables,
        history: Optional[Sequence[ProviderMessage]] = None,
        media_inputs: Optional[MediaInputMap] = None,
    ) -> BoundPrompt:
        # check history for a system message
        history_clean: List[Dict[str, Any]] = []
        if history:
            template_messages_contain_system = any(
                message.role == "system"
                for message in self.messages
                if isinstance(message, TemplateChatMessage)
            )
            history_dict = [convert_provider_message_to_dict(msg) for msg in history]
            for msg in history_dict:
                history_has_system = msg.get("role", None) == "system"
                if history_has_system and template_messages_contain_system:
                    log_freeplay_client_warning(
                        "System message found in history, and prompt template."
                        "Removing system message from the history"
                    )
                else:
                    history_clean.append(msg)

        has_history_placeholder = any(
            isinstance(message, HistoryTemplateMessage) for message in self.messages
        )
        if history and not has_history_placeholder:
            raise FreeplayClientError(
                "History provided for prompt that does not expect history"
            )
        if has_history_placeholder and not history:
            log_freeplay_client_warning(
                "History missing for prompt that expects history"
            )

        bound_messages: List[Dict[str, Any]] = []
        if not media_inputs:
            media_inputs = {}
        for msg in self.messages:
            if isinstance(msg, HistoryTemplateMessage):
                bound_messages.extend(history_clean)
            else:
                media_content = extract_media_content(media_inputs, msg.media_slots)
                content = bind_template_variables(msg.content, variables)

                if media_content:
                    bound_messages.append(
                        {
                            "role": msg.role,
                            "content": [TextContent(text=content), *media_content],
                            "has_media": True,
                        }
                    )
                else:
                    bound_messages.append(
                        {"role": msg.role, "content": content},
                    )

        return BoundPrompt(
            self.prompt_info, bound_messages, self.tool_schema, self.output_schema
        )


class TemplateResolver(ABC):
    @abstractmethod
    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        pass

    @abstractmethod
    def get_prompt(
        self, project_id: str, template_name: str, environment: str
    ) -> PromptTemplate:
        pass

    @abstractmethod
    def get_prompt_version_id(
        self, project_id: str, template_id: str, version_id: str
    ) -> PromptTemplate:
        pass


class FilesystemTemplateResolver(TemplateResolver):
    # If you think you need a change here, be sure to check the server as the translations must match. Once we have
    # all the SDKs and all customers on the new common format, this translation can go away.
    __role_translations = {
        "system": "system",
        "user": "user",
        "assistant": "assistant",
        "Assistant": "assistant",
        "Human": "user",  # Don't think we ever store this, but in case...
    }

    def __init__(self, freeplay_directory: Path):
        FilesystemTemplateResolver.__validate_freeplay_directory(freeplay_directory)
        self.prompts_directory = freeplay_directory / "freeplay" / "prompts"

    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        self.__validate_prompt_directory(project_id, environment)

        directory = self.prompts_directory / project_id / environment
        prompt_file_paths = directory.glob("*.json")

        prompt_list: List[PromptTemplate] = []
        for prompt_file_path in prompt_file_paths:
            json_dom = json.loads(prompt_file_path.read_text())
            prompt_list.append(self.__render_into_v2(json_dom))

        return PromptTemplates(prompt_list)

    def get_prompt(
        self, project_id: str, template_name: str, environment: str
    ) -> PromptTemplate:
        self.__validate_prompt_directory(project_id, environment)

        expected_file: Path = (
            self.prompts_directory / project_id / environment / f"{template_name}.json"
        )

        if not expected_file.exists():
            raise FreeplayClientError(
                f"Could not find prompt with name {template_name} for project "
                f"{project_id} in environment {environment}"
            )

        json_dom = json.loads(expected_file.read_text())
        return self.__render_into_v2(json_dom)

    def get_prompt_version_id(
        self, project_id: str, template_id: str, version_id: str
    ) -> PromptTemplate:
        expected_file: Path = self.prompts_directory / project_id

        if not expected_file.exists():
            raise FreeplayClientError(f"Could not find project id {project_id}")

        # read all files in the project directory
        prompt_file_paths = expected_file.glob("**/*.json")
        # find the file with the matching version id
        for prompt_file_path in prompt_file_paths:
            json_dom = json.loads(prompt_file_path.read_text())
            if json_dom.get("prompt_template_version_id") == version_id:
                return self.__render_into_v2(json_dom)

        raise FreeplayClientError(
            f"Could not find prompt with version id {version_id} for project {project_id}"
        )

    @staticmethod
    def __render_into_v2(json_dom: Dict[str, Any]) -> PromptTemplate:
        format_version = json_dom.get("format_version")

        if format_version and format_version >= 2:
            metadata = json_dom["metadata"]
            flavor_name = metadata.get("flavor")
            model = metadata.get("model")

            return PromptTemplate(
                format_version=format_version,
                prompt_template_id=json_dom.get("prompt_template_id"),  # type: ignore
                prompt_template_version_id=json_dom.get("prompt_template_version_id"),  # type: ignore
                prompt_template_name=json_dom.get("prompt_template_name"),  # type: ignore
                content=FilesystemTemplateResolver.__normalize_messages(
                    json_dom["content"]
                ),
                metadata=PromptTemplateMetadata(
                    provider=FilesystemTemplateResolver.__flavor_to_provider(
                        flavor_name
                    ),
                    flavor=flavor_name,
                    model=model,
                    params=metadata.get("params"),
                    provider_info=metadata.get("provider_info"),
                ),
                project_id=str(json_dom.get("project_id")),
                tool_schema=[
                    ToolSchema(
                        name=schema.get("name"),
                        description=schema.get("description"),
                        parameters=schema.get("parameters"),
                    )
                    for schema in json_dom.get("tool_schema", [])
                ]
                if json_dom.get("tool_schema")
                else None,
                output_schema=json_dom.get("output_schema"),
            )
        else:
            metadata = json_dom["metadata"]

            flavor_name = metadata.get("flavor_name")
            params = metadata.get("params")
            model = params.pop("model") if "model" in params else None

            return PromptTemplate(
                format_version=2,
                prompt_template_id=json_dom.get("prompt_template_id"),  # type: ignore
                prompt_template_version_id=json_dom.get("prompt_template_version_id"),  # type: ignore
                prompt_template_name=json_dom.get("name"),  # type: ignore
                content=FilesystemTemplateResolver.__normalize_messages(
                    json.loads(str(json_dom["content"]))
                ),
                metadata=PromptTemplateMetadata(
                    provider=FilesystemTemplateResolver.__flavor_to_provider(
                        flavor_name
                    ),
                    flavor=flavor_name,
                    model=model,
                    params=params,
                    provider_info=None,
                ),
                project_id=str(json_dom.get("project_id")),
            )

    @staticmethod
    def __normalize_messages(messages: List[Dict[str, Any]]) -> List[TemplateMessage]:
        normalized: List[TemplateMessage] = []
        for message in messages:
            if "kind" in message:
                normalized.append(HistoryTemplateMessage(kind="history"))
            else:
                role = (
                    FilesystemTemplateResolver.__role_translations.get(message["role"])
                    or message["role"]
                )
                raw_media_slots = message.get("media_slots", [])
                media_slots: List[MediaSlot] = (
                    [
                        MediaSlot(
                            type=cast(MediaType, slot["type"]),
                            placeholder_name=cast(str, slot["placeholder_name"]),
                        )
                        for slot in raw_media_slots
                    ]
                    if raw_media_slots
                    else []
                )
                normalized.append(
                    TemplateChatMessage(
                        role=cast(Role, role),
                        content=message["content"],
                        media_slots=media_slots,
                    )
                )
        return normalized

    @staticmethod
    def __validate_freeplay_directory(freeplay_directory: Path) -> None:
        if not freeplay_directory.is_dir():
            raise FreeplayConfigurationError(
                "Path for prompt templates is not a valid directory (%s)"
                % freeplay_directory
            )

        prompts_directory = freeplay_directory / "freeplay" / "prompts"
        if not prompts_directory.is_dir():
            raise FreeplayConfigurationError(
                "Invalid path for prompt templates (%s). "
                "Did not find a freeplay/prompts directory underneath."
                % freeplay_directory
            )

    def __validate_prompt_directory(self, project_id: str, environment: str) -> None:
        maybe_prompt_dir = self.prompts_directory / project_id / environment
        if not maybe_prompt_dir.is_dir():
            raise FreeplayConfigurationError(
                "Could not find prompt template directory for project ID %s and environment %s."
                % (project_id, environment)
            )

    @staticmethod
    def __flavor_to_provider(flavor: str) -> str:
        flavor_provider = {
            "azure_openai_chat": "azure",
            "anthropic_chat": "anthropic",
            "openai_chat": "openai",
            "gemini_chat": "vertex",
        }
        provider = flavor_provider.get(flavor)
        if not provider:
            raise MissingFlavorError(flavor)
        return provider


class APITemplateResolver(TemplateResolver):
    def __init__(self, call_support: CallSupport):
        self.call_support = call_support

    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts(
            project_id=project_id, environment=environment
        )

    def get_prompt(
        self, project_id: str, template_name: str, environment: str
    ) -> PromptTemplate:
        return self.call_support.get_prompt(
            project_id=project_id, template_name=template_name, environment=environment
        )

    def get_prompt_version_id(
        self, project_id: str, template_id: str, version_id: str
    ) -> PromptTemplate:
        return self.call_support.get_prompt_version_id(
            project_id=project_id, template_id=template_id, version_id=version_id
        )


class Prompts:
    def __init__(
        self, call_support: CallSupport, template_resolver: TemplateResolver
    ) -> None:
        self.call_support = call_support
        self.template_resolver = template_resolver

    def get_all(self, project_id: str, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts(
            project_id=project_id, environment=environment
        )

    def get(
        self, project_id: str, template_name: str, environment: str
    ) -> TemplatePrompt:
        prompt = self.template_resolver.get_prompt(
            project_id, template_name, environment
        )

        params = prompt.metadata.params
        model = prompt.metadata.model

        if not model:
            raise FreeplayConfigurationError(
                "Model must be configured in the Freeplay UI. Unable to fulfill request."
            )

        if not prompt.metadata.flavor:
            raise FreeplayConfigurationError(
                "Flavor must be configured in the Freeplay UI. Unable to fulfill request."
            )

        if not prompt.metadata.provider:
            raise FreeplayConfigurationError(
                "Provider must be configured in the Freeplay UI. Unable to fulfill request."
            )

        prompt_info = PromptInfo(
            prompt_template_id=prompt.prompt_template_id,
            prompt_template_version_id=prompt.prompt_template_version_id,
            template_name=prompt.prompt_template_name,
            environment=environment,
            model_parameters=cast(LLMParameters, params) or LLMParameters({}),
            provider=prompt.metadata.provider,
            model=model,
            flavor_name=prompt.metadata.flavor,
            provider_info=prompt.metadata.provider_info,
        )

        return TemplatePrompt(
            prompt_info, prompt.content, prompt.tool_schema, prompt.output_schema
        )

    def get_all_for_environment(self, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts_for_environment(environment=environment)

    def get_by_version_id(
        self, project_id: str, template_id: str, version_id: str
    ) -> TemplatePrompt:
        prompt = self.template_resolver.get_prompt_version_id(
            project_id, template_id, version_id
        )

        params = prompt.metadata.params
        model = prompt.metadata.model

        if not model:
            raise FreeplayConfigurationError(
                "Model must be configured in the Freeplay UI. Unable to fulfill request."
            )

        if not prompt.metadata.flavor:
            raise FreeplayConfigurationError(
                "Flavor must be configured in the Freeplay UI. Unable to fulfill request."
            )

        if not prompt.metadata.provider:
            raise FreeplayConfigurationError(
                "Provider must be configured in the Freeplay UI. Unable to fulfill request."
            )

        prompt_info = PromptInfo(
            prompt_template_id=prompt.prompt_template_id,
            prompt_template_version_id=prompt.prompt_template_version_id,
            template_name=prompt.prompt_template_name,
            environment=prompt.environment if prompt.environment else "",
            model_parameters=cast(LLMParameters, params) or LLMParameters({}),
            provider=prompt.metadata.provider,
            model=model,
            flavor_name=prompt.metadata.flavor,
            provider_info=prompt.metadata.provider_info,
        )

        return TemplatePrompt(
            prompt_info, prompt.content, prompt.tool_schema, prompt.output_schema
        )

    def get_formatted(
        self,
        project_id: str,
        template_name: str,
        environment: str,
        variables: InputVariables,
        history: Optional[Sequence[ProviderMessage]] = None,
        flavor_name: Optional[str] = None,
        media_inputs: Optional[MediaInputMap] = None,
    ) -> FormattedPrompt:
        bound_prompt = self.get(
            project_id=project_id, template_name=template_name, environment=environment
        ).bind(variables=variables, history=history, media_inputs=media_inputs)

        return bound_prompt.format(flavor_name)

    def get_formatted_by_version_id(
        self,
        project_id: str,
        template_id: str,
        version_id: str,
        variables: InputVariables,
        flavor_name: Optional[str] = None,
    ) -> FormattedPrompt:
        bound_prompt = self.get_by_version_id(
            project_id=project_id, template_id=template_id, version_id=version_id
        ).bind(variables=variables)

        return bound_prompt.format(flavor_name)

    def create_version(
        self,
        project_id: str,
        template_name: str,
        template_messages: List[TemplateMessage],
        model: str,
        provider: str,
        version_name: Optional[str] = None,
        version_description: Optional[str] = None,
        llm_parameters: Optional[LLMParameters] = None,
        tool_schema: Optional[List[ToolSchema]] = None,
        environments: Optional[List[str]] = None,
    ) -> Optional[TemplateVersionResponse]:
        return self.call_support.create_version(
            project_id=project_id,
            template_name=template_name,
            template_messages=template_messages,
            model=model,
            provider=provider,
            version_name=version_name,
            version_description=version_description,
            llm_parameters=llm_parameters,
            tool_schema=tool_schema,
            environments=environments,
        )

    def update_version_environments(
        self,
        project_id: str,
        template_id: str,
        template_version_id: str,
        environments: List[str],
    ) -> None:
        self.call_support.update_version_environments(
            project_id=project_id,
            template_id=template_id,
            template_version_id=template_version_id,
            environments=environments,
        )



import os
import unittest
from time import time

import boto3  # type: ignore

from freeplay import CallInfo, Freeplay, RecordPayload
from tests.slow_test_support import slow


class TestBedrockConverse(unittest.TestCase):
    @slow
    def setUp(self) -> None:
        self.freeplay_client = Freeplay(
            freeplay_api_key=os.environ["FREEPLAY_API_KEY"],
            api_base=f"{os.environ['FREEPLAY_API_URL']}/api",
        )
        self.project_id = os.environ["FREEPLAY_PROJECT_ID"]
        self.converse_client = boto3.client(
            service_name="bedrock-runtime",
            region_name="us-west-2",
            aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        )

    @slow
    def test_bedrock_converse(self) -> None:
        start_time = time()
        input_variables = {"query": "What is the capital of the moon?"}
        formatted_prompt = self.freeplay_client.prompts.get_formatted(
            project_id=self.project_id,
            template_name="bedrock-converse",
            environment="latest",
            variables=input_variables,
        )
        response = self.converse_client.converse(
            modelId=formatted_prompt.prompt_info.model,
            messages=formatted_prompt.llm_prompt,
            system=[{"text": formatted_prompt.system_content or ""}],
            inferenceConfig=formatted_prompt.prompt_info.model_parameters,
        )
        output_message = response["output"]["message"]
        response_content = output_message["content"][0]["text"]

        self.assertGreater(len(response_content), 0)

        record_response = self.freeplay_client.recordings.create(
            RecordPayload(
                project_id=self.project_id,
                all_messages=[
                    *formatted_prompt.llm_prompt,
                    output_message,
                ],
                session_info=self.freeplay_client.sessions.create().session_info,
                inputs=input_variables,
                prompt_version_info=formatted_prompt.prompt_info,
                call_info=CallInfo.from_prompt_info(
                    formatted_prompt.prompt_info, start_time, time()
                ),
            )
        )

        self.assertIsNotNone(record_response.completion_id)

    @slow
    def test_bedrock_converse_with_tool_calls(self) -> None:
        start_time = time()
        input_variables = {"query": "What is the latest news in Boulder, CO?"}
        formatted_prompt = self.freeplay_client.prompts.get_formatted(
            project_id=self.project_id,
            template_name="bedrock-converse-tool",
            environment="latest",
            variables=input_variables,
        )
        response = self.converse_client.converse(
            modelId=formatted_prompt.prompt_info.model,
            messages=formatted_prompt.llm_prompt,
            system=[{"text": formatted_prompt.system_content or ""}],
            inferenceConfig=formatted_prompt.prompt_info.model_parameters,
            toolConfig=formatted_prompt.tool_schema,
        )

        output_message = response["output"]["message"]
        tool_call = output_message["content"][0]["toolUse"]
        self.assertGreater(len(tool_call), 0)

        record_response = self.freeplay_client.recordings.create(
            RecordPayload(
                project_id=self.project_id,
                all_messages=[
                    *formatted_prompt.llm_prompt,
                    output_message,
                ],
                session_info=self.freeplay_client.sessions.create().session_info,
                inputs=input_variables,
                prompt_version_info=formatted_prompt.prompt_info,
                call_info=CallInfo.from_prompt_info(
                    formatted_prompt.prompt_info, start_time, time()
                ),
            )
        )

        self.assertIsNotNone(record_response.completion_id)



import os
import time
from typing import cast
from uuid import uuid4

from customer_utils import get_freeplay_thin_client
from openai import OpenAI

fp_client = get_freeplay_thin_client()
openai_client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

project_id = os.environ["FREEPLAY_PROJECT_ID"]
template_prompt = fp_client.prompts.get(
    project_id=project_id, template_name="media-prompt", environment="latest"
)

test_run = fp_client.test_runs.create(
    project_id,
    "media-1",
    include_outputs=True,
    name=f"Media Test Run: {uuid4()}",
    description="Test run with media inputs from Python SDK",
    flavor_name=template_prompt.prompt_info.flavor_name,
)

for test_case in test_run.test_cases:
    print(f"Processing test case: {test_case.id}")

    # Get media inputs from test case
    media_inputs = test_case.media_variables

    formatted_prompt = template_prompt.bind(
        test_case.variables, history=test_case.history, media_inputs=media_inputs
    ).format()

    print(f"Ready for LLM: {formatted_prompt.llm_prompt}")

    start = time.time()
    completion = openai_client.chat.completions.create(
        messages=formatted_prompt.llm_prompt,
        model=formatted_prompt.prompt_info.model,
        **formatted_prompt.prompt_info.model_parameters,
    )
    end = time.time()

    response_content = cast(str, completion.choices[0].message.content)
    print(f"Completion: {response_content}")

    session = fp_client.sessions.create()
    test_run_info = test_run.get_test_run_info(test_case.id)

    # Manually construct messages to avoid JSON serialization issues with media
    all_messages = [
        *formatted_prompt.llm_prompt,
        {"role": "assistant", "content": response_content},
    ]

    # Use the manual recording approach instead of record_results_messages helper
    # since media inputs require special handling
    from freeplay import CallInfo, RecordPayload, ResponseInfo

    call_info = CallInfo(
        formatted_prompt.prompt_info.provider,
        model=formatted_prompt.prompt_info.model,
        start_time=start,
        end_time=end,
        model_parameters=formatted_prompt.prompt_info.model_parameters,
    )

    response_info = ResponseInfo(
        is_complete=completion.choices[0].finish_reason == "stop"
    )

    fp_client.recordings.create(
        RecordPayload(
            project_id=project_id,
            all_messages=all_messages,
            session_info=session.session_info,
            inputs=test_case.variables,
            media_inputs=media_inputs,
            prompt_version_info=formatted_prompt.prompt_info,
            call_info=call_info,
            response_info=response_info,
            test_run_info=test_run_info,
            eval_results={
                "has_media": media_inputs is not None,
                "media_count": len(media_inputs or {}),
                "response_length": len(response_content) if response_content else 0,
            },
        )
    )

# wait 5 sec and get the results
time.sleep(5)
results = fp_client.test_runs.get(project_id, test_run.test_run_id)
print(results.test_run_id)
print(results.name)
print(results.description)
print(results.summary_statistics)



import copy
import json
import os
import time
from typing import Any, Dict, List, Union
import boto3  # type: ignore

from freeplay import CallInfo, Freeplay, RecordPayload


def add_numbers(numbers: List[int]) -> int:
    return sum(numbers)


def multiple_numbers(numbers: List[int]) -> int:
    result = 1
    for number in numbers:
        result *= number
    return result


def subtract_two_numbers(a: int, b: int) -> int:
    return a - b


def divide_two_numbers(a: int, b: int) -> float:
    return a / b


def execute_function(func_name: str, args: Dict[str, Any]) -> Union[int, float]:
    if func_name == "add_numbers":
        return add_numbers(args["numbers"])
    elif func_name == "multiple_numbers":
        return multiple_numbers(args["numbers"])
    elif func_name == "subtract_two_numbers":
        return subtract_two_numbers(args["a"], args["b"])
    elif func_name == "divide_two_numbers":
        return divide_two_numbers(args["a"], args["b"])
    else:
        raise Exception("Function not found")


toolsSpec = [
    {
        "toolSpec": {
            "name": "add_numbers",
            "description": "Add a list of numbers",
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {
                        "numbers": {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "List of numbers to add",
                        }
                    },
                    "required": ["numbers"],
                }
            },
        }
    },
    {
        "toolSpec": {
            "name": "multiple_numbers",
            "description": "Multiply a list of numbers",
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {
                        "numbers": {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "List of numbers to multiply",
                        }
                    },
                    "required": ["numbers"],
                }
            },
        }
    },
    {
        "toolSpec": {
            "name": "subtract_two_numbers",
            "description": "Subtract two numbers",
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {
                        "a": {"type": "integer", "description": "First number"},
                        "b": {"type": "integer", "description": "Second number"},
                    },
                    "required": ["a", "b"],
                }
            },
        }
    },
    {
        "toolSpec": {
            "name": "divide_two_numbers",
            "description": "Divide two numbers",
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {
                        "a": {"type": "integer", "description": "First number"},
                        "b": {"type": "integer", "description": "Second number"},
                    },
                    "required": ["a", "b"],
                }
            },
        }
    },
]


fp_client = Freeplay(
    api_base=f"{os.environ.get('FREEPLAY_API_URL')}/api",
    freeplay_api_key=os.environ.get("FREEPLAY_API_KEY") or "",
)
aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID") or ""
aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY") or ""
if not aws_access_key_id or not aws_secret_access_key:
    raise Exception("AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be set")
converse_client = boto3.client(  # type: ignore
    service_name="bedrock-runtime",
    region_name="us-east-1",
    aws_access_key_id=aws_access_key_id,
    aws_secret_access_key=aws_secret_access_key,
)
project_id: str = os.environ["FREEPLAY_PROJECT_ID"]

equation = "2x + 5 = 10"
prompt_vars = {"equation": equation}

try:
    formatted_prompt = fp_client.prompts.get_formatted(
        project_id=project_id,
        template_name="nova_tool_call",
        environment="latest",
        variables=prompt_vars,
    )
except Exception as e:
    print(f"Error details: {e}")
    print(f"Project ID: {project_id}")
    print("Template: nova_tool_call")
    print("Environment: latest")
    raise

session = fp_client.sessions.create()
trace = session.create_trace(input=equation)

print(f"Using model: {formatted_prompt.prompt_info.model}")
print(f"Template: {formatted_prompt.prompt_info.template_name}")

# Initialize history with the formatted prompt messages
history = copy.deepcopy(formatted_prompt.llm_prompt)

finish_reason = None
while finish_reason not in ["end_turn", "stop"]:
    s = time.time()
    print(f"Model: {formatted_prompt.prompt_info.model}")

    # Make the Bedrock Converse call
    response = converse_client.converse(  # type: ignore
        modelId=formatted_prompt.prompt_info.model,
        messages=history,
        system=[{"text": formatted_prompt.system_content or ""}],
        inferenceConfig=formatted_prompt.prompt_info.model_parameters,
        toolConfig={"tools": toolsSpec},
    )
    e = time.time()
    print(f"Response: {json.dumps(response, indent=2, default=str)}")

    output_message = response["output"]["message"]  # type: ignore
    finish_reason = response["stopReason"]  # type: ignore

    print(f"Stop reason: {finish_reason}")

    # Check for tool calls
    if finish_reason == "tool_use":
        # Find the toolUse in content (may not be first item due to thinking text)
        tool_use = None
        for content_item in output_message["content"]:  # type: ignore
            if "toolUse" in content_item:
                tool_use = content_item["toolUse"]  # type: ignore
                break

        if not tool_use:
            raise Exception("No toolUse found in response")
        tool_name = tool_use["name"]  # type: ignore
        tool_input = tool_use["input"]  # type: ignore
        tool_id = tool_use["toolUseId"]  # type: ignore

        print(
            f"Executing function {tool_name} with args {json.dumps(tool_input, indent=2)}"
        )
        result: Union[int, float] = execute_function(tool_name, tool_input)  # type: ignore
        print(f"Result: {result}")
        print("\n")

        # Add the full assistant response to history (includes thinking + toolUse)
        print("=== Adding assistant message to history ===")
        print(f"Assistant message: {json.dumps(output_message, indent=2, default=str)}")
        history.append(output_message)

        # Add the tool response to history
        tool_result_message: Dict[str, Any] = {
            "role": "user",
            "content": [
                {
                    "toolResult": {
                        "toolUseId": tool_id,
                        "content": [{"text": str(result)}],
                    }
                }
            ],
        }
        print("\n=== Adding tool result to history ===")
        print(
            f"Tool result message: {json.dumps(tool_result_message, indent=2, default=str)}"
        )
        history.append(tool_result_message)

        # Record the tool call to freeplay
        print("\n=== Recording to Freeplay ===")
        print(f"History length: {len(history)}")
        print("Full history:")
        print(json.dumps(history, indent=2, default=str))
        print("\nTool schema being sent:")
        print(json.dumps(toolsSpec, indent=2, default=str))

        payload = RecordPayload(
            project_id=project_id,
            all_messages=history,
            inputs=prompt_vars,
            session_info=session.session_info,
            trace_info=trace,
            prompt_version_info=formatted_prompt.prompt_info,
            call_info=CallInfo.from_prompt_info(
                formatted_prompt.prompt_info, start_time=s, end_time=e
            ),
        )

        print("\n=== Payload details ===")
        print(f"Project ID: {project_id}")
        print(f"Session ID: {session.session_info.session_id}")
        print("Is complete: False")
        print(f"Function call: {tool_name}({json.dumps(tool_input)})")

        try:
            fp_client.recordings.create(payload)
            print("\n✓ Successfully recorded to Freeplay")
        except Exception as e:
            print(f"\n✗ Error recording to Freeplay: {e}")
            raise
    else:
        # Final response
        content = output_message["content"][0]["text"]  # type: ignore
        print("=== Solution ===")
        print(content)  # type: ignore
        print("\n")

        # Add the final response to history
        print("=== Adding final response to history ===")
        print(f"Final message: {json.dumps(output_message, indent=2, default=str)}")
        history.append(output_message)

        # Record the final response to freeplay
        print("\n=== Recording final response to Freeplay ===")
        print(f"History length: {len(history)}")
        print("Full history:")
        print(json.dumps(history, indent=2, default=str))

        payload = RecordPayload(
            project_id=project_id,
            all_messages=history,
            inputs=prompt_vars,
            session_info=session.session_info,
            trace_info=trace,
            prompt_version_info=formatted_prompt.prompt_info,
            call_info=CallInfo.from_prompt_info(
                formatted_prompt.prompt_info, start_time=s, end_time=e
            ),
        )

        print("\n=== Payload details ===")
        print(f"Project ID: {project_id}")
        print(f"Session ID: {session.session_info.session_id}")
        print("Is complete: True")

        try:
            fp_client.recordings.create(payload)
            print("\n✓ Successfully recorded to Freeplay")
        except Exception as e:
            print(f"\n✗ Error recording to Freeplay: {e}")
            raise

        trace.record_output(project_id=project_id, output=content)  # type: ignore



import importlib.metadata
import json
import platform
from typing import Dict, Union, Optional, Any

import pystache  # type: ignore

from .errors import FreeplayError, FreeplayConfigurationError
from .model import InputVariables


# Validate that the variables are of the correct type, and do not include functions, dates, classes or None values.
def all_valid(obj: Any) -> bool:
    if isinstance(obj, (int, str, bool, float)):
        return True
    elif isinstance(obj, list):
        return all(all_valid(item) for item in obj)
    elif isinstance(obj, dict):
        return all(
            isinstance(key, str) and all_valid(value) for key, value in obj.items()
        )
    else:
        return False


class StandardPystache(pystache.Renderer):  # type: ignore
    def __init__(self) -> None:
        super().__init__(escape=lambda s: s)

    def str_coerce(self, val: Any) -> str:
        if isinstance(val, dict) or isinstance(val, list):
            # We hide spacing after punctuation so that the templating is the same across all SDKs.
            return json.dumps(val, separators=(",", ":"))
        return str(val)


def bind_template_variables(template: str, variables: InputVariables) -> str:
    if not all_valid(variables):
        raise FreeplayError(
            "Variables must be a string, number, bool, or a possibly nested"
            " list or dict of strings, numbers and booleans."
        )

    # When rendering mustache, do not escape HTML special characters.
    rendered: str = StandardPystache().render(template, variables)
    return rendered


def check_all_values_string_or_number(
    metadata: Optional[Dict[str, Union[str, int, float]]],
) -> None:
    if metadata:
        for key, value in metadata.items():
            if not isinstance(value, (str, int, float)):
                raise FreeplayConfigurationError(
                    f"Invalid value for key {key}: Value must be a string or number."
                )


def build_request_header(api_key: str) -> Dict[str, str]:
    return {"Authorization": f"Bearer {api_key}", "User-Agent": get_user_agent()}


def get_user_agent() -> str:
    sdk_name = "Freeplay"
    sdk_version = importlib.metadata.version("Freeplay")
    language = "Python"
    language_version = platform.python_version()
    os_name = platform.system()
    os_version = platform.release()

    # Output format
    # Freeplay/0.2.30 (Python/3.11.4; Darwin/23.2.0)
    return f"{sdk_name}/{sdk_version} ({language}/{language_version}; {os_name}/{os_version})"


# Recursively convert Pydantic models, lists, and dicts to dict compatible format -- used to allow us to accept
# provider message shapes (usually generated types) or the default {'content': ..., 'role': ...} shape.
def convert_provider_message_to_dict(obj: Any) -> Any:
    """
    Convert provider message objects to dictionaries.
    For Vertex AI objects, automatically converts to camelCase.
    """
    # List of possible raw attribute names in Vertex AI objects
    vertex_raw_attrs = [
        "_raw_content",  # For Content objects
        "_raw_tool",  # For Tool objects
        "_raw_message",  # For message objects
        "_raw_candidate",  # For Candidate objects
        "_raw_response",  # For response objects
        "_raw_function_declaration",  # For FunctionDeclaration
        "_raw_generation_config",  # For GenerationConfig
        "_pb",  # Generic protobuf attribute
    ]

    # Check for Vertex AI objects with raw protobuf attributes
    for attr_name in vertex_raw_attrs:
        if hasattr(obj, attr_name):
            raw_obj = getattr(obj, attr_name)
            if raw_obj is not None:
                try:
                    # Use the metaclass to_dict with camelCase conversion
                    return type(
                        raw_obj
                    ).to_dict(
                        raw_obj,
                        preserving_proto_field_name=False,  # camelCase
                        use_integers_for_enums=False,  # Keep as strings (we'll lowercase them)
                        including_default_value_fields=False,  # Exclude defaults
                    )
                except:  # noqa: E722
                    # If we can't convert, continue to the next attribute
                    pass

    # For non-Vertex AI objects, use their standard to_dict methods
    if hasattr(obj, "to_dict") and callable(getattr(obj, "to_dict")):
        # Regular to_dict (for Vertex AI wrappers without _raw_* attributes)
        return obj.to_dict()
    elif hasattr(obj, "model_dump"):
        # Pydantic v2
        return obj.model_dump(mode="json")
    elif hasattr(obj, "dict"):
        # Pydantic v1
        return obj.dict(encode_json=True)
    elif isinstance(obj, dict):
        # Handle dictionaries recursively
        return {k: convert_provider_message_to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        # Handle lists recursively
        return [convert_provider_message_to_dict(item) for item in obj]

    # Return as-is for primitive types
    return obj



import json
import os
import sys
from pathlib import Path
from stat import S_IREAD, S_IRGRP, S_IROTH, S_IWUSR

import click

from freeplay.errors import FreeplayClientError, FreeplayServerError
from freeplay import Freeplay
from freeplay.support import PromptTemplates, PromptTemplate, PromptTemplateEncoder


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option("--project-id", required=True, help="The Freeplay project ID.")
@click.option(
    "--environment",
    required=True,
    help="The environment from which the prompts will be pulled.",
)
@click.option(
    "--output-dir", required=True, help="The directory where the prompts will be saved."
)
def download(project_id: str, environment: str, output_dir: str) -> None:
    if "FREEPLAY_API_KEY" not in os.environ:
        print(
            "FREEPLAY_API_KEY is not set. It is required to run the freeplay command.",
            file=sys.stderr,
        )
        exit(4)

    if "FREEPLAY_SUBDOMAIN" not in os.environ:
        print(
            "FREEPLAY_SUBDOMAIN is not set. It is required to run the freeplay command.",
            file=sys.stderr,
        )
        exit(4)

    FREEPLAY_API_KEY = os.environ["FREEPLAY_API_KEY"]
    freeplay_api_url = f"https://{os.environ['FREEPLAY_SUBDOMAIN']}.freeplay.ai/api"

    if "FREEPLAY_API_URL" in os.environ:
        freeplay_api_url = f"{os.environ['FREEPLAY_API_URL']}/api"
        click.echo(
            "Using URL override for Freeplay specified in the FREEPLAY_API_URL environment variable"
        )

    click.echo(
        f"Downloading prompts for project {project_id}, environment {environment}, "
        f"to directory {output_dir} from {freeplay_api_url}"
    )

    fp_client = Freeplay(freeplay_api_key=FREEPLAY_API_KEY, api_base=freeplay_api_url)

    try:
        prompts: PromptTemplates = fp_client.prompts.get_all(
            project_id, environment=environment
        )
        click.echo(f"Found {len(prompts.prompt_templates)} prompt templates")

        for prompt in prompts.prompt_templates:
            __write_single_file(environment, output_dir, prompt)
    except FreeplayClientError as e:
        print(
            f"Error downloading templates: {e}.\nIs your project ID correct?",
            file=sys.stderr,
        )
        exit(1)
    except FreeplayServerError as e:
        print(
            f"Error on Freeplay's servers downloading templates: {e}.\nTry again after a short wait.",
            file=sys.stderr,
        )
        exit(2)
    except Exception as e:
        print(f"Error downloading templates: {e}", file=sys.stderr)
        exit(3)


@cli.command()
@click.option(
    "--environment",
    required=True,
    help="The environment from which the prompts will be pulled.",
    default="latest",
)
@click.option(
    "--output-dir", required=True, help="The directory where the prompts will be saved."
)
def download_all(environment: str, output_dir: str) -> None:
    if "FREEPLAY_API_KEY" not in os.environ:
        print(
            "FREEPLAY_API_KEY is not set. It is required to run the freeplay command.",
            file=sys.stderr,
        )
        exit(4)

    if "FREEPLAY_SUBDOMAIN" not in os.environ:
        print(
            "FREEPLAY_SUBDOMAIN is not set. It is required to run the freeplay command.",
            file=sys.stderr,
        )
        exit(4)

    FREEPLAY_API_KEY = os.environ["FREEPLAY_API_KEY"]
    freeplay_api_url = f"https://{os.environ['FREEPLAY_SUBDOMAIN']}.freeplay.ai/api"

    if "FREEPLAY_API_URL" in os.environ:
        freeplay_api_url = f"{os.environ['FREEPLAY_API_URL']}/api"
        click.echo(
            "Using URL override for Freeplay specified in the FREEPLAY_API_URL environment variable"
        )

    click.echo(
        f"Downloading prompts for environment {environment}, "
        f"to directory {output_dir} from {freeplay_api_url}"
    )

    fp_client = Freeplay(freeplay_api_key=FREEPLAY_API_KEY, api_base=freeplay_api_url)

    try:
        prompts: PromptTemplates = fp_client.prompts.get_all_for_environment(
            environment=environment
        )
        click.echo(f"Found {len(prompts.prompt_templates)} prompt templates")

        for prompt in prompts.prompt_templates:
            __write_single_file(environment, output_dir, prompt)
    except FreeplayClientError as e:
        print(
            f"Error downloading templates: {e}.\nIs your project ID correct?",
            file=sys.stderr,
        )
        exit(1)
    except FreeplayServerError as e:
        print(
            f"Error on Freeplay's servers downloading templates: {e}.\nTry again after a short wait.",
            file=sys.stderr,
        )
        exit(2)
    except Exception as e:
        print(f"Error downloading templates: {e}", file=sys.stderr)
        exit(3)


def __write_single_file(
    environment: str, output_dir: str, prompt: PromptTemplate
) -> None:
    directory = __root_dir(environment, output_dir, prompt.project_id)
    basename = f"{prompt.prompt_template_name}"
    prompt_path = directory / f"{basename}.json"
    click.echo("Writing prompt file: %s" % prompt_path)

    # Make sure it's owner writable if it already exists
    if prompt_path.is_file():
        os.chmod(prompt_path, S_IWUSR | S_IREAD)

    with prompt_path.open(mode="w") as f:
        f.write(json.dumps(prompt, sort_keys=True, indent=4, cls=PromptTemplateEncoder))
        f.write("\n")

    # Make the file read-only to discourage local changes
    os.chmod(prompt_path, S_IREAD | S_IRGRP | S_IROTH)


def __root_dir(environment: str, output_dir: str, project_id: str) -> Path:
    directory = (
        Path(output_dir).resolve() / "freeplay" / "prompts" / project_id / environment
    )
    os.makedirs(directory, exist_ok=True)
    return directory



import os
import time

from openai import OpenAI

from freeplay import CallInfo, Freeplay, RecordPayload, ResponseInfo

fpclient = Freeplay(
    freeplay_api_key=os.environ["FREEPLAY_API_KEY"],
    api_base=f"{os.environ['FREEPLAY_API_URL']}/api",
)
client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

input_variables = {"question": "What's in this image?"}

image_url = "https://upload.wikimedia.org/wikipedia/commons/a/a7/Camponotus_flavomarginatus_ant.jpg"

formatted_prompt = fpclient.prompts.get_formatted(
    project_id=os.environ["FREEPLAY_PROJECT_ID"],
    template_name="media",
    environment="latest",
    variables=input_variables,
    history=[
        {
            "role": "user",
            "content": [
                {"type": "image_url", "image_url": {"url": image_url, "detail": "auto"}}
            ],
        }
    ],
)

start = time.time()
completion = client.chat.completions.create(
    messages=formatted_prompt.llm_prompt,
    model=formatted_prompt.prompt_info.model,
    **formatted_prompt.prompt_info.model_parameters,
)
end = time.time()

print("Completion:", completion.choices[0].message.content)

session = fpclient.sessions.create()
all_messages = formatted_prompt.all_messages(completion.choices[0].message)
call_info = CallInfo.from_prompt_info(formatted_prompt.prompt_info, start, end)
response_info = ResponseInfo(is_complete=completion.choices[0].finish_reason == "stop")

record_response = fpclient.recordings.create(
    RecordPayload(
        project_id=os.environ["FREEPLAY_PROJECT_ID"],
        all_messages=all_messages,
        session_info=session.session_info,
        inputs=input_variables,
        prompt_version_info=formatted_prompt.prompt_info,
        call_info=call_info,
        tool_schema=formatted_prompt.tool_schema,
        response_info=response_info,
    )
)

print(f"Recorded completion ID: {record_response.completion_id}")
